
#include "assert.h"

#include "buffer.h"
#include "heap.h"

#define BUFFER_SIZE_MAX     ((uint16_t)INT16_MAX)

#define FREESIZE_MAX( _buffer ) ((_buffer->size != 0 ) ? (_buffer->size - 1) : _buffer->size)  // Handling the "empty-shell" buffer

typedef enum
{
    BUFFER_TYPE_STANDARD = 0,
    BUFFER_TYPE_SHELL,
} buffer_type_t;

struct buffer
{
    buffer_type_t type;

    uint8_t * data;
    uint16_t size;          /* NOTE: The REAL maximum size of the buffer will be size-1 due to the round buffer nature */
    uint16_t read_pos;
    uint16_t write_pos;

    uint16_t loadsize;
    uint16_t freesize;

    uint16_t freesize_limit;
};

/* Prototypes */
static ret_code_t _buffer_config( buffer_t * _buffer, const buffer_conf_t * _conf );

/*
 *  It is allowed to initiate a "shell" buffer by configuring data == NULL and size == 0.
 *
 *  SUCH SHELL MUST NOT BE USED FOR ANY KIND OF PROCESSING UNTIL AFTER IT HAS BEEN FURTHER INITIALIZED,
 *  E.G. BY MAKING A CLONE WITH buffer_make_clone.
 */

static ret_code_t _buffer_init( buffer_t * _buffer, const buffer_conf_t * _conf )
{
    ret_code_t result = NRF_SUCCESS;
    buffer_conf_t config = *_conf;

    /* If the data is not specified in the _config and the size is demanded as non-zero, try to allocate it */
    if ( config.data == NULL && config.size != 0 )
    {
        config.data = heap_alloc( config.size );
    }

    result = _buffer_config( _buffer, &config );
    EXIT_IF_ERR( result, "_buffer_config failed" );

    _buffer->type = BUFFER_TYPE_STANDARD;

_EXIT:
    return result;
}

ret_code_t buffer_init( buffer_t ** __buffer, const buffer_conf_t * _conf )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    *__buffer = heap_alloc( sizeof( buffer_t ) );

    result = _buffer_init( *__buffer, _conf );
    EXIT_IF_ERR( result, "_buffer_init failed" );

_EXIT:
    return result;
}

ret_code_t buffer_init_shell( buffer_t ** __buffer )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    buffer_conf_t conf =
    {
        .data = NULL,
        .size = 0,
    };

    result = buffer_init( __buffer, &conf );
    EXIT_IF_ERR(result, "Buffer init shell - buffer_init failed.")

    (*__buffer)->type = BUFFER_TYPE_SHELL;

_EXIT:
    return result;
}

static ret_code_t _buffer_config( buffer_t * _buffer, const buffer_conf_t * _conf )
{
    ret_code_t result = NRF_SUCCESS;

    ASSERT_RULE( _conf->size <= BUFFER_SIZE_MAX, "buffer size exceeds its maximum value" );

    /* Initialize the buffer */
    _buffer->data = _conf->data;
    _buffer->size = _conf->size;

    buffer_clear( _buffer );

    /* Initialize the freesize limit */
    buffer_reset_freesize_limit( _buffer );

    return result;
}

ret_code_t buffer_config( buffer_t * _buffer, const buffer_conf_t * _conf )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    ASSERT_RULE( _buffer->type == BUFFER_TYPE_SHELL, "Buffer config - Only the Shell buffer may change its base configuration." );

    result = _buffer_config( _buffer, _conf );
    EXIT_IF_ERR( result, "Buffer config - _buffer_config failed" );

_EXIT:
    return result;
}

void buffer_clear( buffer_t * _buffer )
{
    _buffer->read_pos = 0;
    _buffer->write_pos = 0;

    _buffer->loadsize = 0;
    _buffer->freesize = _buffer->freesize_limit;
}

void buffer_make_clone( buffer_t * _clone, const buffer_t * _image )
{
    ASSERT_RULE( _clone->type == BUFFER_TYPE_SHELL, "Buffer make clone - Only the Shell buffer may be used as a clone." );

    ASSERT_RULE( _clone != NULL && _image != NULL, "Buffer make clone - clone or image is NULL." );

    buffer_type_t clone_type = _clone->type;

    *_clone = *_image;

    _clone->type = clone_type;
}

void buffer_set_freesize_limit( buffer_t * _buffer, uint16_t freesize_limit )
{
    ASSERT_RULE( freesize_limit <= FREESIZE_MAX( _buffer ), "Buffer set freesize limit - demanded limit is higher than the maximum freesize." );

    /* Check "no change" situation */
    if ( freesize_limit == _buffer->freesize_limit )
    {
        return;
    }

    ASSERT_RULE( freesize_limit >= _buffer->loadsize, "Buffer set freesize limit - loadsize already exceedes the demanded freesize limit." );


    _buffer->freesize_limit = freesize_limit;
    _buffer->freesize = _buffer->freesize_limit - _buffer->loadsize;
}

void buffer_reset_freesize_limit( buffer_t * _buffer )
{
    buffer_set_freesize_limit( _buffer, FREESIZE_MAX( _buffer ) );
}

buffer_conf_t buffer_get_conf( buffer_t * _buffer )
{
    buffer_conf_t conf;

    conf.data = _buffer->data;
    conf.size = _buffer->size;

    return conf;
}

uint16_t buffer_get_read_pos( buffer_t * _buffer )
{
    return _buffer->read_pos;
}

uint16_t buffer_get_write_pos( buffer_t * _buffer )
{
    return _buffer->write_pos;
}

uint16_t buffer_get_loadsize( buffer_t * _buffer )
{
    return _buffer->loadsize;
}

uint16_t buffer_get_freesize( buffer_t * _buffer )
{
    return _buffer->freesize;
}

static INLINE uint16_t _get_max_loadsize( buffer_t * _buffer )
{
    return _buffer->size - 1;
}

uint16_t buffer_get_max_loadsize( buffer_t * _buffer )
{
    return _get_max_loadsize( _buffer );
}

void buffer_update_read_pos( buffer_t * _buffer, int16_t len )
{
    if ( len == 0 )
    {
        return;
    }
    else if ( len > 0 )
    {
        ASSERT_RULE( ( len <= _buffer->loadsize ), "The positive buffer position update length must not be greater than the buffer->loadsize value" );

        _buffer->read_pos = ( uint16_t )( ( int16_t )_buffer->read_pos + len );
    }
    else if ( len < 0 )
    {
        ASSERT_RULE( ( (-len) <= _buffer->freesize ), "The negative buffer position update length must not be greater than the buffer->freesize value" );

        _buffer->read_pos = ( uint16_t )( (int16_t)(_buffer->read_pos + _buffer->size) + len);
    }

    if ( _buffer->read_pos >= _buffer->size )
    {
        _buffer->read_pos -= _buffer->size;
    }

    /* Update buffer sizes */
    _buffer->loadsize = ( uint16_t )( ( int16_t )_buffer->loadsize - len );
    _buffer->freesize = ( uint16_t )( ( int16_t )_buffer->freesize + len );
}

void buffer_update_write_pos( buffer_t * _buffer, int16_t len )
{
    if ( len == 0 ) {
        return;
    }
    else if ( len > 0 )
    {
        ASSERT_RULE( ( len <= _buffer->freesize ), "The positive buffer position update length must not be greater than the buffer->freesize value" );

        _buffer->write_pos = ( uint16_t )( ( int16_t )_buffer->write_pos + len );
    }
    else if ( len < 0 )
    {
        ASSERT_RULE( ( ( -len ) <= _buffer->loadsize ), "The negative buffer position update length must not be greater than the buffer->loadsize value" );

        _buffer->write_pos = ( uint16_t )( ( int16_t )( _buffer->write_pos + _buffer->size ) + len );
    }

    if ( _buffer->write_pos >= _buffer->size )
    {
        _buffer->write_pos -= _buffer->size;
    }

    /* Update buffer sizes */
    _buffer->loadsize = ( uint16_t )( ( int16_t )_buffer->loadsize + len );
    _buffer->freesize = ( uint16_t )( ( int16_t )_buffer->freesize - len );
}

static uint16_t _get_pos_offset( uint16_t buffer_size, uint16_t old_pos, uint16_t new_pos )
{
    if ( new_pos >= old_pos )
    {
        return new_pos - old_pos;
    }
    else
    {
        return buffer_size + new_pos - old_pos;
    }
}

static uint16_t _get_read_pos_offset( buffer_t * _buffer, uint16_t new_read_pos )
{
    return _get_pos_offset( _buffer->size, _buffer->read_pos, new_read_pos );
}

static uint16_t _get_write_pos_offset( buffer_t * _buffer, uint16_t new_write_pos )
{
    return _get_pos_offset( _buffer->size, _buffer->write_pos, new_write_pos );
}

/* We use the buffer_update_read_pos to maintain the buffer consistency */
void buffer_set_read_pos_safe( buffer_t * _buffer, uint16_t read_pos )
{
    uint16_t read_pos_offset;

    if ( read_pos == _buffer->read_pos )
    {
        return;
    }

    ASSERT_RULE( ( read_pos < _buffer->size ), "New buffer read position must be lower than the buffer->size value" );

    /* Get read position offset from the new read_pos value */
    read_pos_offset = _get_read_pos_offset( _buffer, read_pos );

    /* Update the read position of the buffer */
    buffer_update_read_pos( _buffer, read_pos_offset );
}

/* We use the buffer_update_write_pos to maintain the buffer consistency */
void buffer_set_write_pos_safe( buffer_t * _buffer, uint16_t write_pos )
{
    uint16_t write_pos_offset;

    if ( write_pos == _buffer->write_pos ) {
        return;
    }

    ASSERT_RULE( ( write_pos < _buffer->size ), "New buffer write position must be lower than the buffer->size value" );

    /* Get write position offset from the new write_pos value */
    write_pos_offset = _get_write_pos_offset( _buffer, write_pos );

    /* Update the write position of the buffer */
    buffer_update_write_pos( _buffer, write_pos_offset );
}

/* Copy data between two buffers. Specifies the position in the source "from buffer" from which to start the copy  */
ret_code_t buffer_copy( buffer_t * _buffer_from, uint16_t from_position, buffer_t * _buffer_to, uint16_t len )
{
    /* Check the sizes */
    if ( len == 0 )
    {
        return NRF_SUCCESS;
    }
    else if( (len > buffer_get_max_loadsize( _buffer_from ) ) || (len > _buffer_to->freesize ) )
    {
        return NRF_ERROR_INTERNAL;
    }

    /* Add data from _buffer_from to _buffer_to */
    if ( from_position + len > _buffer_from->size )
    {
        uint16_t first_part_len = _buffer_from->size - from_position;

        buffer_add( _buffer_to, &_buffer_from->data[ from_position ], first_part_len );
        buffer_add( _buffer_to, &_buffer_from->data[ 0 ], len - first_part_len );
    }
    else
    {
        buffer_add( _buffer_to, &_buffer_from->data[ from_position ], len );
    }

    return NRF_SUCCESS;
}

/* Move data between two circular buffers, i.e. copy and move from read position */
ret_code_t buffer_move( buffer_t * _buffer_from, buffer_t * _buffer_to, uint16_t len )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    if( (len > _buffer_from->loadsize) )
    {
        return NRF_ERROR_INTERNAL;
    }

    /* Copy the data */
    result = buffer_copy( _buffer_from, _buffer_from->read_pos, _buffer_to, len );
    EXIT_IF_ERR( result, "buffer_move failed" );

    buffer_update_read_pos( _buffer_from, len );

_EXIT:
    return NRF_SUCCESS;
}

/* Move any data between two circular buffers, i.e. copy and move from read position */
ret_code_t buffer_move_any( buffer_t * _buffer_from, buffer_t * _buffer_to, uint16_t * len_moved )
{
    ret_code_t result = NRF_ERROR_INTERNAL;
    uint16_t buffer_to_freesize;
    uint16_t len;

    /* Determine the maximum length of data which can be moved */
    len = _buffer_from->loadsize;
    buffer_to_freesize = _buffer_to->freesize;

    if ( len > buffer_to_freesize )
    {
        len = buffer_to_freesize;
    }

    /* Move the data */
    result = buffer_move( _buffer_from, _buffer_to, len );
    EXIT_IF_ERR( result, "buffer_move failed" );

    /* Store the resulting length of the moved data */
    if ( len_moved != NULL )
    {
        *len_moved = len;
    }

_EXIT:
    return result;
}

/*
 * Function for inserting data anywhere in the buffer space. This is only the data write operation without any change of the
 * buffer r/w positions.
 */
static void _insert_data( buffer_t * _buffer, const uint8_t data[], uint16_t write_pos, uint16_t len )
{
    /*
     * NOTE: Assert could be here. However, there are currently other checks in the superior functions which ensure the buffer does not get overflown
     */

    /* Optimize for speed */
    if ( len == 1 )
    {
        _buffer->data[write_pos] = data[0];
    }
    else if ( len != 0 )
    {
        /* Copy data */
        if ( write_pos + len > _buffer->size )
        {
            uint16_t first_part_len = _buffer->size - write_pos;

            memcpy( _buffer->data + write_pos, data, first_part_len );
            memcpy( _buffer->data, data + first_part_len, len - first_part_len );
        }
        else
        {
            memcpy( _buffer->data + write_pos, data, len );
        }
    }
}

/*
 * Warning: use this carefully as this function might overwrite valid data in the buffer. Read more in the description of _insert_data.
 */
void buffer_insert( buffer_t * _buffer, const uint8_t data[], uint16_t write_pos, uint16_t len )
{
    ASSERT_RULE( len < buffer_get_max_loadsize( _buffer ), "The length of data must not be more than the maximum loadsize of the buffer" );

    _insert_data( _buffer, data, write_pos, len );
}

ret_code_t buffer_add( buffer_t * _buffer, const uint8_t data[], uint16_t len )
{
    /* Check the size */
    if ( len > _buffer->freesize )
    {
        return NRF_ERROR_INTERNAL;
    }
    else
    {
        _insert_data( _buffer, data, _buffer->write_pos, len );
    }

    /* Update write position */
    buffer_update_write_pos( _buffer, len );

    return NRF_SUCCESS;
}

ret_code_t buffer_get_from_position( buffer_t * _buffer,
    uint16_t from_position,
    uint8_t * data,
    uint16_t len )
{
    ASSERT_RULE( ( data != NULL ), "buffer_get target is NULL!" );

    /* Check the size */
    if ( len > _buffer->loadsize )
    {
        return NRF_ERROR_INTERNAL;
    }
    /* Optimizing for speed */
    else if ( len == 1 )
    {
        *data = _buffer->data[ from_position ];
    }
    else if ( len != 0 )
    {
        if ( from_position + len > _buffer->size )
        {
            memcpy( data, _buffer->data + from_position, _buffer->size - from_position );
            memcpy( data + _buffer->size - from_position, _buffer->data, len - _buffer->size + from_position );
        }
        else
        {
            memcpy( data, _buffer->data + from_position, len );
        }
    }

    return NRF_SUCCESS;
}

ret_code_t buffer_get( buffer_t * _buffer,
    uint8_t * data,
    uint16_t len )
{
    return buffer_get_from_position( _buffer, _buffer->read_pos, data, len );
}

ret_code_t buffer_get_and_discard( buffer_t * _buffer,
    uint8_t * data,
    uint16_t len )
{
    ret_code_t result = NRF_ERROR_INTERNAL;

    result = buffer_get( _buffer, data, len );
    EXIT_IF_ERR( result, "buffer_get failed" );

    buffer_update_read_pos( _buffer, len );

_EXIT:
    return result;
}

uint8_t * buffer_get_load_space_pointer( buffer_t * _buffer, uint16_t offset )
{
    uint16_t offset_read_pos = buffer_calc_new_pos( _buffer, _buffer->read_pos, offset );

    return &_buffer->data[ offset_read_pos ];
}

uint8_t * buffer_get_free_space_pointer( buffer_t * _buffer )
{
    return &_buffer->data[ _buffer->write_pos ];
}

/* Returns the size either "from read_pos to write_pos" or "from read_pos to the end of data space"  */
uint16_t buffer_get_load_space_line_size( const buffer_t * _buffer, uint16_t offset )
{
    uint16_t offset_read_pos = buffer_calc_new_pos( _buffer, _buffer->read_pos, offset );

    if ( _buffer->write_pos >= offset_read_pos )
    {
        return _buffer->write_pos - offset_read_pos;
    }
    else
    {
        return _buffer->size - offset_read_pos;
    }
}

/* Returns the size either "from write_pos to read_pos" or "from write_pos to the end of data space"  */
uint16_t buffer_get_free_space_line_size( const buffer_t * _buffer )
{
    if ( _buffer->read_pos > _buffer->write_pos )
    {
        /* Subtract 1 to prevent write position potentially jumping at the read position and thus "clearing the buffer" */
        return _buffer->read_pos - _buffer->write_pos - 1;
    }
    /* Special case: If the read position is at 0, the write position must not wrap around at the end of the buffer data space */
    else if ( _buffer->read_pos == 0 )
    {
        return _buffer->size - _buffer->write_pos - 1;
    }
    else
    {
        return _buffer->size - _buffer->write_pos;
    }
}

/* Calculate new position with respect to the buffer and its size, respectively. Do not affect the original buffer. */
uint16_t buffer_calc_new_pos( const buffer_t * _buffer, uint16_t old_pos, int16_t offset )
{
    uint16_t new_pos;

    if ( offset == 0 )
    {
        return old_pos;
    }
    else if ( offset > 0 )
    {
        ASSERT_RULE( ( offset < _buffer->size ), "The positive buffer position offset must be smaller than the buffer->size value" );

        new_pos = ( uint16_t )( ( int16_t )old_pos + offset );
    }
    else if ( offset < 0 )
    {
        ASSERT_RULE( ( (-offset) < _buffer->size ), "The negative buffer position offset must be smaller than the buffer->size value" );

        new_pos = ( uint16_t )( (int16_t)(old_pos + _buffer->size) + offset);
    }

    if ( new_pos >= _buffer->size )
    {
        new_pos -= _buffer->size;
    }

    return new_pos;
}

void buffer_print_data( const buffer_t * _buffer )
{
    if( _buffer->read_pos + _buffer->loadsize > _buffer->size )
    {
        DEBUG_PRINT_DATA( &_buffer->data[_buffer->read_pos], _buffer->size - _buffer->read_pos );
        DEBUG_PRINTF(" ");
        DEBUG_PRINT_DATA( &_buffer->data[0], _buffer->write_pos );
    }
    else
    {
        DEBUG_PRINT_DATA( &_buffer->data[_buffer->read_pos], _buffer->loadsize );
    }
}
